# -*- coding: utf-8 -*-

import yaml
import logging
import logging.config

from artixds.domain import DiscountSystem, DiscountCampaign
from subsystem import Subsystem
from dao import loadDiscountSystem
from dao import getDiscountUpdateTime
import dao
from document import Document
import ujson

try:
    logging.config.fileConfig('/linuxcash/cash/discountsystems/logging.properties')
except Exception as e:
    print(u'logger configure error: `%s`' % e)
    logging.basicConfig()

session = dao.createSession()
discountSystem, secondDiscountSystem, bonusEarnSystem, bonusLimitSystem = DiscountSystem(), DiscountSystem(), DiscountSystem(), DiscountSystem()
time_update = loadDiscountSystem(session, discountSystem, secondDiscountSystem, bonusEarnSystem, bonusLimitSystem)

log = logging.getLogger("su.artix.dm")

if log.isEnabledFor(logging.DEBUG):
    log.debug(u'Первая дисконтная система:')
    for n, line in enumerate(filter(lambda x: x.rstrip(), yaml.dump(discountSystem).split('\n'))):
        log.debug(u'%4d: %s' % (n, line.rstrip()))
    log.debug(u'Вторая дисконтная система:')
    for n, line in enumerate(filter(lambda x: x.rstrip(), yaml.dump(secondDiscountSystem).split('\n'))):
        log.debug(u'%4d: %s' % (n, line.rstrip()))
    log.debug(u'Начисление бонусов:')
    for n, line in enumerate(filter(lambda x: x.rstrip(), yaml.dump(bonusEarnSystem).split('\n'))):
        log.debug(u'%4d: %s' % (n, line.rstrip()))
    log.debug(u'Списание бонусов:')
    for n, line in enumerate(filter(lambda x: x.rstrip(), yaml.dump(bonusLimitSystem).split('\n'))):
        log.debug(u'%4d: %s' % (n, line.rstrip()))


def refreshSystems():
    global time_update
    global manualCampaigns
    old_time_update = time_update
    log.debug(u'Обновление дисконтной системы')
    time_update = loadDiscountSystem(session, discountSystem, secondDiscountSystem, bonusEarnSystem, bonusLimitSystem, time_update)
    return Subsystem().encode({'refreshSystems': old_time_update != time_update})


def isUpToDate():
    global time_update
    global session
    updateTime, newSession = getDiscountUpdateTime(session)
    session = newSession
    return Subsystem().encode({'isUpToDate': time_update is not None and updateTime == time_update})


def calculateDiscountImpact(documentJsonString={}, priority=None, checkPriority=False, qualifiers=[]):
    return calculateDiscountImpactForSubsystem(discountSystem, documentJsonString, priority, checkPriority, qualifiers)


def calculateSecondDiscountImpact(documentJsonString={}, priority=None, checkPriority=False, qualifiers=[]):
    return calculateDiscountImpactForSubsystem(secondDiscountSystem, documentJsonString, priority, checkPriority, qualifiers)


def calculateDiscountImpactForSubsystem(ds, documentJsonString, priority, checkPriority, qualifiers):
    if qualifiers and (set(qualifiers) & ds.qualifiers) != set(qualifiers):
        log.debug(u'Дисконтная система не содержит спецификаторов %s' % str(qualifiers))
        # TODO: очень плохое решение - нужно возвращать из дисконта словарь, но это долго делать
        return Subsystem.encode([[], [], [], [], [], [], [], [], []])

    filters = [lambda campaign: set(qualifiers) == (campaign.qualifiers)]
    subsystem = Subsystem(ds)
    subsystem.setQualifiersFilters(filters)
    return subsystem.getJsonImpacts(Document(ujson.loads(documentJsonString)), priority, checkPriority)


def calculateBonusEarningsImpact(documentJsonString, qualifiers=[]):
    if qualifiers and (set(qualifiers) & bonusEarnSystem.qualifiers) != set(qualifiers):
        log.debug(u'Бонусная система не содержит спецификаторов %s' % str(qualifiers))
        # TODO: очень плохое решение - нужно возвращать из дисконта словарь, но это долго делать
        return Subsystem.encode([[], [], [], [], [], [], [], [], []])

    filters = [lambda campaign: set(qualifiers) == (campaign.qualifiers)]
    subsystem = Subsystem(bonusEarnSystem)
    subsystem.setQualifiersFilters(filters)
    return subsystem.getJsonImpacts(Document(ujson.loads(documentJsonString)))


def calculateBonusLimitImpact(documentJsonString):
    return Subsystem(bonusLimitSystem).getJsonImpacts(Document(ujson.loads(documentJsonString)))


def getCampaigns():
    return Subsystem().encode(discountSystem.manualCampaigns)


def getSecondDiscountSystemCampaigns():
    return Subsystem().encode(secondDiscountSystem.manualCampaigns)


def getGroupAndCampaignsPriorities():
    return Subsystem().encode(discountSystem.groupsAndCampaignsPriorities)


def getSecondDiscountSystemGroupAndCampaignsPriorities():
    return Subsystem().encode(secondDiscountSystem.groupsAndCampaignsPriorities)


def calculateLifeTimeLimit(data=''):
    try:
        campaign = yaml.load(data, Loader=yaml.CLoader)
        campaign.calcAssetDates()
        return Subsystem().encode({
            "assetBeginDate": campaign.assetBeginDate,
            "assetEndDate": campaign.assetEndDate,
            "id": campaign.id,
            "weight": campaign.weight
        })
    except Exception as e:
        log.warning('%s', str(e))
    return {}
