# -*- coding: utf-8 -*-

import copy
from datetime import datetime
from artixds.domain import DiscountCampaign, DiscountCampaignGroup, Coupon, Counter, CounterChanged
from artixds.checkwrapper import CheckWrapper
from artixds.domain import aggregateImpact, ImpactGroup, InteractionType, ImpactType
from artixds.conditionfunctionplugins.conditionplugin import conditionMemory
from artixds.ratefunctionplugins.rateplugin import rateMemory
import jsonpickle
import re
import time
import logging
from artixds.conditionfunction import ConditionFunction
from artixds.objectdiscount import CheckObject

logging.basicConfig(level=logging.DEBUG, format='[%(levelname)s]-%(name)s: %(message)s \nFile "%(pathname)s", line %(lineno)d')
log = logging.getLogger("Subsystem")
checkNumerator = 0
firstAdd = []


def getCurrentCheck(check, campaignGroup):
    """ Return check or its copy """
    if campaignGroup.interactionType == InteractionType.ALL.value:
        global checkNumerator
        if campaignGroup.id in firstAdd and check.isModified:
            tmp = checkNumerator
            checkNumerator = tmp + 1
            check.id = checkNumerator
        else:
            firstAdd.append(campaignGroup.id)
        check.isModified = False
        return check
    return CheckWrapper().copy(check)


class EasyPureImpact(object):
    pass


class EasyStructureImpact(object):
    def __init__(self, posnum, qty, sum):
        self.posnum = posnum
        self.qty = qty
        self.sum = sum

    def __repr__(self):
        return "posnum=%s;qty=%s;sum=%s" % (self.posnum, self.qty, self.sum)


class DiscountCampaignImpact(object):
    """
    Create impacts at the DiscountCampaign
    """
    def __init__(self, campaign, check, durationCampaigns):
        self.campaign = campaign
        self.check = check
        self.durationCampaigns = durationCampaigns

    def getImpacts(self):
        """
        Get impacts for discount campaign
        """
        campaign = self.campaign
        interaction = campaign.interactionType
        allImpacts = []
        self.check.newGroup()

        if interaction == InteractionType.LAST.value and not campaign.isDiscountListReverse():
            campaign.setDiscountListInReverse()
        for discount in campaign.discounts:
            if discount.isActive():
                timeStartDiscount = time.time()
                impactGroup = DiscountImpacts(discount).getImpact(getCurrentCheck(self.check, campaign))
                allImpacts.append(impactGroup)
                if impactGroup.getImpacts() and interaction in [InteractionType.FIRST.value, InteractionType.LAST.value]:
                    break
                duration = time.time() - timeStartDiscount
                self.durationCampaigns['%s(%s)' % (campaign.name, campaign.id)]['%s(%s)' % (discount.name, discount.id)] = round(duration, 6)

        result = aggregateImpact(allImpacts, interaction)
        return result


class DiscountContext(object):
    """
    Контекст скидки. При принятии решения о применении каждой новой скидки
    создаётся контекст скидки, который существует до момента образования
    воздействия
    """
    def __init__(self):
        self.cards = []  # список карт, удовлетворяющий условию скидки, в порядке добавления карт в документ
        self.coupons = []
        self.counters = []
        self.posDependent = []
        self.posIndependent = []

    def __repr__(self):
        return repr(self.cards)


class DiscountImpacts(object):
    """
    Create impacts at the Discount
    """
    def __init__(self, discount):
        """
        discount - Discount
        posDependent - list of conditions, depending on the position
        posIndependent = list of conditions, independing on the position
        """
        self.discount = discount

    def selectImpacts(self, obj, check, discount, discountContext, getImpactFunction):
        demoModeFunctions = [self.discount.getLegacyImpact, self.discount.getSimpleImpact]
        if not check.isFullMode() and getImpactFunction not in demoModeFunctions:
            log.debug(u"Недоступно в демонстрационном режиме")
            return None, False

        specific, total = discount.isConditionsTrue(obj, check, discountContext)
        if total:
            return None, True
        if specific:
            impact = getImpactFunction(check, obj, discountContext)
            return impact.getImpacts(), total
        else:
            return None, False

    def getImpact(self, check):
        """
        Get impacts for discount
        return single ImpactGroup
        """
        discount = self.discount
        discount.showCashTextToConsultant = getattr(discount, 'showCashTextToConsultant', False)
        discount.clientDisplayText = getattr(discount, 'clientDisplayText', None)
        discount.notPosConditionTrue = False

        try:
            value = ""
            for kitDiscountRate in discount.discountRate.items:
                value = value + repr(kitDiscountRate.discountRate.value) + " " + repr(kitDiscountRate.discountRate.type) + ", "
        except AttributeError:
            value = repr(discount.discountRate.value) + " " + repr(discount.discountRate.type)

        log.debug(u"    Скидка [%s]", discount.name)
        impactGroup = ImpactGroup()
        checkObjExist, posObjExist, kitObjExist, legObjExist, kitNewObjExist, simpleObjExist = discount.presenceObjects()
        discountContext = DiscountContext()

        # формируем списки условий, зависящих и не зависящих от позиций
        contentPosConditions = ConditionFunction().posConditionList
        for conditItem in discount.conditions:
            posDependentCondition = list(filter(lambda x: x in conditItem.condition, contentPosConditions))
            if posDependentCondition:
                discountContext.posDependent.append(conditItem)
            else:
                discountContext.posIndependent.append(conditItem)

        if posObjExist:
            for obj in check.getGoodsList():
                imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getPosImpact)
                if res:
                    break
                if imp:
                    impactGroup.addImpacts(imp)

        if checkObjExist:
            obj = check.getCheckObject()
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getCheckImpact)
            if imp:
                impactGroup.addImpacts(imp)

        obj = check.createKit()
        if kitObjExist and discount.isValidKitObject(obj):
            log.debug(u"        Найден набор")
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getKitImpact)
            if imp:
                impactGroup.addImpacts(imp)

        if kitNewObjExist:
            log.debug(u"        Создан новый набор")
            obj = check.getCheckObject()
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getKitNewImpact)
            if imp:
                impactGroup.addImpacts(imp)

        if legObjExist:
            obj = check.getCheckObject()
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getLegacyImpact)
            if imp:
                impactGroup.addImpacts(imp)

        if simpleObjExist:
            obj = check.getCheckObject()
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getSimpleImpact)
            if imp:
                impactGroup.addImpacts(imp)

        log.debug(u"        Cтавка [%s]", value)

        impactGroup.check = check

        if not impactGroup.getImpacts():
            return impactGroup

        # добавляем сообщение, если скидка применена
        if log.isEnabledFor(logging.DEBUG):
            if discount.cashText:
                log.debug(u'    Сообщение на экран кассира: "%s"' % discount.cashText)
            if discount.clientText:
                log.debug(u'    Сообщение для печати в чеке: "%s"' % discount.clientText)
            if discount.clientDisplayText:
                log.debug(u'    Сообщение на экран клиента: "%s"' % discount.clientDisplayText)

        # запомним первую попавшуюся карту для того, чтобы использовать в макросах
        cardForMacro = discountContext.cards[0] if discountContext.cards else None

        if discount.cashText and discount.cashText.strip():
            if check.checkDict.get("cashtype") != "SSCO" or discount.showCashTextToConsultant:
                impactGroup.screenMessage.append(self.__prepareMessage(discount.cashText, cardForMacro, impactGroup))
        if discount.clientText and discount.clientText.strip():
            impactGroup.receiptMessage.append(self.__prepareMessage(discount.clientText, cardForMacro, impactGroup))
        if discount.clientDisplayText and discount.clientDisplayText.strip():
            if check.checkDict.get("cashtype") == "SSCO":
                impactGroup.clientDisplayMessage.append(self.__prepareMessage(discount.clientDisplayText, cardForMacro, impactGroup))
        if discount.reports:
            impactGroup.reports.extend(discount.reports)
        if discount.coupons:
            for c in discount.coupons:
                c.report = self.__prepareMessage(c.report, cardForMacro, impactGroup)
                c.campaignId = discount.campaign.id
                c.campaignName = discount.campaign.name
                impactGroup.coupons.append(c)

        # обработка расчетных функций для счетчиков дисконта
        if discount.counters:
            counters = {}
            countersDestType = {}
            for counter in discount.counters:
                counter.rate.checkWrapper = check

                # 0 - счетчики привязанные к карте, 1 - свободные счетчики
                destType = getattr(counter, 'destinationType', 0)
                if (destType == 0 and check.checkDict["cardrecords"]) or destType == 1:
                    if counter.position:
                        for position in check.getGoodsList():
                            counterId = self.__streval(counter.counterId, check=check.checkDict, position=position.data)
                            counterValue = counter.rate.getAbsoluteValue(position, check.checkDict)
                            counters[counterId] = counters.get(counterId, 0) + counterValue
                            countersDestType[counterId] = destType
                    else:
                        counterId = self.__streval(counter.counterId, check=check.checkDict)
                        counterValue = counter.rate.getAbsoluteValue(CheckObject(check.checkDict), check.checkDict)
                        counters[counterId] = counters.get(counterId, 0) + counterValue
                        countersDestType[counterId] = destType

                delattr(counter.rate, "checkWrapper")

            for counterId in sorted(counters):
                if abs(counters[counterId]) > 0.001:
                    log.debug(u"Изменение счетчика дисконта '%s': %s" % (counterId, counters[counterId]))
                    impactGroup.counters.append(CounterChanged(counterId, counters[counterId], discount.campaign.id,
                            discount.campaign.qualifiers, countersDestType[counterId],
                            [{'codeinternal': card.get('codeinternal')} for card in discountContext.cards]))

        return impactGroup

    def __prepareMessage(self, message, card=None, impactGroup=None):
        replaceBlocks = []
        for matchObject in re.finditer(r'\${(.+?)}', message):
            expression = matchObject.group(1)
            try:
                expression, format = expression.split(':', 1) if ':' in expression else [expression, None]
                value = self.__eval(expression, impactGroup.check.checkDict, card, impactGroup)
                if str(type(value)) in ['<class \'document.QDateTime\'>', '<class \'document.QDate\'>']:
                    if value.isValid():
                        value = value.toString(format) if format else value.toString()
                    else:
                        value = "---"
                else:
                    if value and not (value == "None" or value == ""):
                        value = ''.join(['{0:', format, '}']).format(value) if format else value
                    else:
                        value = "---"
                replaceBlocks.append((matchObject.group(0), value))
            except Exception as e:
                log.error("Prepare message. Error in eval call: '%s'" % expression)
                log.exception(e)
                replaceBlocks.append((matchObject.group(0), "---"))

        # заменим полученные блоки
        preparedMessage = message
        for oldString, newString in replaceBlocks:
            preparedMessage = preparedMessage.replace(oldString, newString)
        return preparedMessage

    @staticmethod
    def __streval(expression, check=None, position=None):
        """
        Метод для выполнения выражения, которое может быть строкой, выполнение которой не ожидалось, или содержать строки
        Пример: 'counter1' - без изменений; '"777_" + position["code"]' - получаем 777_КодПозиции
        """
        try:
            scope = dict(globals(), **locals())
            return str(eval("%s" % expression, scope))
        except Exception:
            return str(expression)

    def __eval(self, expression, check, card, impact):
        scope = dict(globals(), **locals())
        return eval(expression, scope)


class Subsystem(object):
    """
    TODO: переписать на C++
    """
    def __init__(self, discountSystem=None):
        if discountSystem is not None:
            self.rootCampaignGroup = discountSystem.rootCampaignGroup
            self.discountSystemNumber = discountSystem.number
        self.qualifiersFilters = [lambda campaign: campaign.qualifiers == set()]
        self.durationCampaigns = {}

    @staticmethod
    def encode(data):
        return jsonpickle.encode(data, unpicklable=False)

    def setQualifiersFilters(self, filters):
        self.qualifiersFilters = filters

    def getJsonImpacts(self, document, priority=None, checkPriority=False):
        # очищаем время выполнения дисконта и кэш с результатами вычислений условий и расчетных ставок
        conditionMemory.clear()
        rateMemory.clear()

        return Subsystem.encode(list(self.__quick(document, priority, checkPriority)))

    # создание структурного воздействия для передачи на Кассу
    def __quick(self, document, priority, checkPriority):
        if not self.rootCampaignGroup:
            return [], [], [], [], [], [], [], [], []

        impactGroup = self.__getGroupImpacts(CheckWrapper(document), self.rootCampaignGroup, priority, checkPriority)

        self.__printDuration()

        if not impactGroup:
            return [], [], [], [], [], [], [], [], []

        modify = {}
        parentSubs = {}
        for x in list(filter(lambda x: x.type == ImpactType.ADD.value, impactGroup.getImpacts())):
            log.debug("ADD %d parent %d" % (x.impactObject.positionNumber, x.parentPositionNumber))
            parentPosition = parentSubs.get(x.parentPositionNumber, x.parentPositionNumber)
            parentSubs[x.impactObject.positionNumber] = x.parentPositionNumber
            if parentPosition not in modify:
                modify[parentPosition] = []
            mod = EasyStructureImpact()
            mod.posnum = x.impactObject.positionNumber
            mod.qty = x.qty
            mod.sum = impactGroup.check.getPositionByNumber(mod.posnum)["sumb"]
            modify[parentPosition].append(mod)

        for x in list(filter(lambda x: x.type == ImpactType.BRANCH.value, impactGroup.getImpacts())):
            modify[x.impactObject.positionNumber] = [EasyStructureImpact(det.positionNumber, det.quantity, det.sumb) for det in x.branchDetails]

        impacts = []
        for x in list(filter(lambda x: x.type == ImpactType.PURE.value, impactGroup.getImpacts())):
            impact = EasyPureImpact()
            impact.discountMode = ("%s" % x.discountMode).lower()
            impact.absoluteValue = x.absoluteValue
            impact.positionNumber = x.impactObject.positionNumber
            impact.quantity = float(x.qty) if x.qty else x.qty

            impact.discountRate = copy.copy(x.discountRate if x.discountRate else x.discount.discountRate)

            impact.discount = copy.copy(x.discount)
            impact.campaign = copy.copy(x.discount.campaign)

            impact.cards = [{"codeinternal": x.cards[0]['codeinternal']}] if x.cards else []
            impact.coupons = [{"number": x.coupons[0]['number']}] if x.coupons else []

            impact.campaign.id = x.campaign.id
            impact.campaign.name = x.campaign.name
            if isinstance(impact.campaign, DiscountCampaign):
                impact.campaign.calcAssetDates(document)
                for c in impactGroup.coupons:
                    if c.campaignId == impact.campaign.id:
                        c.beginDate = impact.campaign.assetBeginDate
                        c.endDate = impact.campaign.assetEndDate

                for c in impactGroup.counters:
                    if c.campaignId == impact.campaign.id:
                        c.beginDate = impact.campaign.assetBeginDate
                        c.endDate = impact.campaign.assetEndDate

            self.__removeUnusedProperties(impact)

            impacts.append(impact)

        uniteImpacts = {}
        for x in list(filter(lambda x: x.type == ImpactType.UNITE.value, impactGroup.getImpacts())):
            uniteImpacts[x.impactObject.positionNumber] = [EasyStructureImpact(pos, quant, None) for pos, quant in x.uniteDetails.items()]

        coupons_res = copy.deepcopy(impactGroup.coupons)
        counters_res = copy.deepcopy(impactGroup.counters)

        return impacts, modify, impactGroup.screenMessage, impactGroup.receiptMessage, impactGroup.clientDisplayMessage,\
               uniteImpacts, coupons_res, counters_res, impactGroup.reports

    def __getGroupImpacts(self, check, campaignGroup, priority, checkPriority):
        """ Получение воздействий для группы акций """
        allImpacts = []
        log.debug(u"Группа акций [%s]", campaignGroup.name)
        check.newGroup()

        """
        objects - objects of discountCampaigns[] + objects of childGroups[]
        """
        objects = campaignGroup.discountCampaigns + campaignGroup.childGroups
        interaction = campaignGroup.interactionType
        # InteractionType.LAST сводится к InteractionType.FIRST сменой порядка применения акций(условий) на обратный
        if interaction == InteractionType.LAST.value:
            objects.reverse()

        for obj in objects:
            if isinstance(obj, DiscountCampaign):
                timeStartCampaign = time.time()
                currentCheck = getCurrentCheck(check, campaignGroup)
                campaign = obj
                self.durationCampaigns['%s(%s)' % (campaign.name, campaign.id)] = {}
                if not self.__checkCampaignQualifiers(campaign):
                    continue

                if not checkPriority or (checkPriority and str(campaign.priority) == str(priority)):
                    if campaign.isValidDate(check.getDataTimeBeg()) and campaign.isActive():
                        activatedCampaigns = []
                        if 'activatedcampaigns' in check.checkDict:
                            activatedCampaigns = check.checkDict["activatedcampaigns"]
                        if not campaign.manualActivate or (str(self.discountSystemNumber) in activatedCampaigns and campaign.id in activatedCampaigns[str(self.discountSystemNumber)]):
                            cashTags = []
                            if checkPriority:
                                log.debug(u"Акция [%s] с приоритетом %s" % (campaign.name, str(campaign.priority)))
                            log.debug(u"Акция [%s], проверка меток:", campaign.name)
                            if 'context' in check.checkDict and 'shoplabels' in check.checkDict["context"]:
                                cashTags = check.checkDict["context"]["shoplabels"].split(",")
                            if campaign.getCheckedTags() is None:
                                log.debug(u"У акции нет меток")
                            if campaign.getCheckedTags() is None or campaign.isValidTag(cashTags):
                                log.debug(u"Акция активна")
                                impactList = DiscountCampaignImpact(campaign, currentCheck, self.durationCampaigns).getImpacts()
                                allImpacts.append(impactList)
                                if impactList.getImpacts() and interaction in [InteractionType.FIRST.value, InteractionType.LAST.value]:
                                    break
                        else:
                            log.debug("Campaign [%s] for manual activation is not activated", campaign.name)
                    else:
                        log.debug(u"Акция [%s] не активна", campaign.name)
                duration = time.time() - timeStartCampaign
                self.durationCampaigns['%s(%s)' % (campaign.name, campaign.id)]["total"] = round(duration, 6)
            elif isinstance(obj, DiscountCampaignGroup):
                if checkPriority:
                    log.debug(u"Группа акций [%s] с приоритетом %s" % (obj.name, str(obj.priority)))
                currentCheck = getCurrentCheck(check, campaignGroup)
                impact = self.__getGroupImpacts(currentCheck, obj, priority, checkPriority)
                if impact and impact.check:
                    allImpacts.append(impact)
                    if impact.getImpacts() and interaction in [InteractionType.FIRST.value, InteractionType.LAST.value]:
                        break

        # отметим позиции, на которые применено воздействие
        impacts = aggregateImpact(allImpacts, campaignGroup.interactionType)
        if impacts and len(impacts.getImpacts()) > 0:
            log.debug(u"Примененные воздействия:")
            for impact in list(filter(lambda x: x.type == ImpactType.PURE.value, impacts.getImpacts())):
                if impact.impactObject.positionNumber != 0:
                    discount = (u"от скидки [" + impact.discount.name + u"] с id [" + repr(impact.discount.id) +
                                u"] из акции [" + impact.discount.campaign.name + u"]")
                    if impact.impactObject.positionNumber == -1:
                        log.debug(u"На чек: %s %s", impact.absoluteValue, discount)
                    else:
                        log.debug(u"На позицию под номером %s (код товара: [%s]): %s %s", impact.impactObject.positionNumber,
                                  impact.impactObject.code, impact.absoluteValue, discount)
                    check.markPositionByDiscount(impact.impactObject.positionNumber)

        return impacts

    def __checkCampaignQualifiers(self, campaign):
        for campaignFilter in self.qualifiersFilters:
            if not campaignFilter(campaign):
                log.debug(u"Акция [%s] не прошла проверку спецификаторов", campaign.name)
                return False
        return True

    def __removeUnusedProperties(self, impact):
        """
        Удаление неиспользуемых значений из воздействия
        """
        for prop in ['campaign', 'objects', 'conditions', 'discountRate', 'clientDisplayText', 'clientText', 'cashText', 'active', 'notPosConditionTrue']:
            impact.discount.__dict__.pop(prop, None)

        if impact.discount.counters == []:
            impact.discount.__dict__.pop('counters', None)

        for prop in ['campaign', 'parentGroup', 'discounts', 'parent', 'beginBonusesIntervalCondition',
                     'beginBonusesIntervalValue', 'beginDateWithBeginCurrentMetric', 'beginDateWithBeginNextMetric',
                     'beginBonusesIntervalMetric', 'endBonusesIntervalMetric', 'endBonusesIntervalCondition',
                     'endBonusesIntervalValue', 'beginBonusesDateCalculate', 'endBonusesDateCalculate',
                     'beginRangeDate', 'endRangeDate', 'checkedTags', 'tagsIsAllowed', 'tagsValid', 'beginDate', 'endDate',
                     'showCashTextToConsultant', 'clientDisplayText', 'clientText', 'cashText', 'conditions', 'minPriceIgnored',
                     'objects', 'discountRate', 'active', 'notPosConditionTrue', 'interactionType', 'allowedTags',
                     'forbiddenTags', 'discountReverse', 'counters']:

            impact.campaign.__dict__.pop(prop, None)

        for prop in ['checkWrapper', 'items', 'mode']:
            impact.discountRate.__dict__.pop(prop, None)

    def __printDuration(self):
        log.debug(u"Время расчета акций и скидок")
        for campaignName, discItem in self.durationCampaigns.items():
            log.debug(u"(%.6f сек) - %s" % (discItem["total"] if "total" in discItem else 0, campaignName))
            for discName, discTime in discItem.items():
                if discName != "total":
                    log.debug(u"        (%.6f сек) - %s" % (discTime, discName))
